import re
import warnings
import logging
from pathlib import Path

from primeight.column import CassandraColumn
from primeight.generators import Generators


logger = logging.getLogger(__name__)


class Parser:

    _required_fields = ['version', 'keyspace', 'name', 'columns', 'query']
    _optional_fields = ['generated_columns', 'split', 'lifetime', 'backup']

    _recognized_splits = ['day', 'week', 'month', 'year']

    _required_query_fields = ['required']
    _optional_query_fields = ['optional', 'order', 'description']
    _recognized_query_required = ['time', 'space', 'id']

    _recognized_orders = ['asc', 'desc']

    _recognized_types = [k.lower() for k in CassandraColumn.HANDLE_TO_CASSANDRA_TYPE.keys()]

    @staticmethod
    def parse(path: str or Path) -> dict:
        pass

    @staticmethod
    def is_valid_name(name: str) -> bool:
        """Validate if a name is valid."""
        valid_name_re = r'^[a-z]+([a-z]*[0-9]*_?)*$'
        cassandra_reserved_keywords = [
            'ADD', 'AGGREGATE', 'ALL', 'ALLOW', 'ALTER', 'AND', 'ANY', 'APPLY',
            'AS', 'ASC', 'ASCII', 'AUTHORIZE', 'BATCH', 'BEGIN', 'BIGINT',
            'BLOB', 'BOOLEAN', 'BY', 'CLUSTERING', 'COLUMNFAMILY', 'COMPACT',
            'CONSISTENCY', 'COUNT', 'COUNTER', 'CREATE', 'CUSTOM', 'DECIMAL',
            'DELETE', 'DESC', 'DISTINCT', 'DOUBLE', 'DROP', 'EACH_QUORUM',
            'ENTRIES', 'EXISTS', 'FILTERING', 'FLOAT', 'FROM', 'FROZEN', 'FULL',
            'GRANT', 'IF', 'IN', 'INDEX', 'INET', 'INFINITY', 'INSERT', 'INT',
            'INTO', 'KEY', 'KEYSPACE', 'KEYSPACES', 'LEVEL', 'LIMIT', 'LIST',
            'LOCAL_ONE', 'LOCAL_QUORUM', 'MAP', 'MATERIALIZED', 'MODIFY', 'NAN',
            'NORECURSIVE', 'NOSUPERUSER', 'NOT', 'OF', 'ON', 'ONE', 'ORDER',
            'PARTITION', 'PASSWORD', 'PER', 'PERMISSION', 'PERMISSIONS',
            'PRIMARY', 'QUORUM', 'RENAME', 'REVOKE', 'SCHEMA', 'SELECT', 'SET',
            'STATIC', 'STORAGE', 'SUPERUSER', 'TABLE', 'TEXT', 'TIME',
            'TIMESTAMP', 'TIMEUUID', 'THREE', 'TO', 'TOKEN', 'TRUNCATE', 'TTL',
            'TUPLE', 'TWO', 'TYPE', 'UNLOGGED', 'UPDATE', 'USE', 'USER',
            'USERS', 'USING', 'UUID', 'VALUES', 'VARCHAR', 'VARINT', 'VIEW',
            'WHERE', 'WITH', 'WRITETIME'
        ]

        if re.match(valid_name_re, name) is None:
            return False

        if name.upper() in cassandra_reserved_keywords:
            return False

        return True

    @staticmethod
    def is_valid_version(version: str) -> bool:
        """Validate if a version is valid."""
        valid_version_re = r'[0-9](.[0-9]+){0,2}'

        if re.match(valid_version_re, version) is None:
            return False

        return True

    @staticmethod
    def is_valid_config(content: dict) -> bool:
        """Validates Yaml has the correct format.
        Throws SyntaxError if Yaml has a wrong format.

        We are very strict with the Yaml format.
        So, wrong format may be caused by:
        * missing required field
        * unrecognized field
        * column name with invalid characters
        * unrecognized column type
        * unrecognized generator
        * generated columns dependent on undeclared columns
        * unrecognized backup strategy
        * unrecognized split strategy
        * non integer lifetime
        * negative lifetime
        * query name with invalid characters
        * query unrecognized field
        * query required field keys not valid
        * undeclared columns as query required or optional columns

        #TODO: improve docstring
        """
        # Validate that all required fields are present.
        missing_fields = [f for f in Parser._required_fields if f not in content]
        if len(missing_fields) > 0:
            missing_fields_str = ', '.join(missing_fields)
            raise SyntaxError(f"Missing required fields: {missing_fields_str}")

        # Validate that the table has a valid name.
        if not Parser.is_valid_name(content['name']):
            raise SyntaxError(
                f"Table name '{content['name']}' is invalid")

        # Validate that, if present, keyspace has a valid name.
        if 'keyspace' in content and not Parser.is_valid_name(content['keyspace']):
            raise SyntaxError(
                f"Keyspace '{content['keyspace']}' is invalid")

        # Validate version.
        if not Parser.is_valid_version(content['version']):
            raise SyntaxError(
                f"Table version '{content['version']}' is invalid")

        all_fields = Parser._required_fields + Parser._optional_fields

        # Validate that no unrecognized field exist.
        unrecognized_fields = [f for f in content if f not in all_fields]
        if len(unrecognized_fields) > 0:
            unrecognized_fields_str = ', '.join(unrecognized_fields)
            raise SyntaxError(f"Unrecognized fields: {unrecognized_fields_str}")

        # Validate that all columns have recognized types.
        if len('columns') == 0:
            raise SyntaxError("columns list is empty")

        # Validate keyspace partition.
        # If no keyspace is declared, then a client_id column is required,
        # and vice-versa.
        if ('keyspace' not in content
                and 'client_id' not in content['columns']):
            raise SyntaxError("column 'client_id' is required in "
                              "when keyspace is not defined.")

        # Validate that client_id column has a type.
        elif ('client_id' in content['columns']
                and 'type' not in content['columns']['client_id']):
            raise SyntaxError(f"Missing type for column client_id")

        # Validate that the client_id column has type text.
        elif ('client_id' in content['columns']
                and content['columns']['client_id']['type'] != 'text'):
            raise SyntaxError(
                f"column client_id has wrong type, must be text")

        for name, col_content in content['columns'].items():

            # Validate that the columns has type.
            if 'type' not in col_content:
                raise SyntaxError(f"Missing type for column {name}")
            if col_content['type'] is None:
                raise SyntaxError(f"column {name} type is empty")

            t = col_content['type'] \
                .lower() \
                .replace(' ', '') \
                .replace('list<', '') \
                .replace('tuple<', '') \
                .replace('set<', '') \
                .strip('>')
            # Validate that the column type is recognized.
            if ',' in t:
                t_parts = t.split(',')
                for tp in t_parts:
                    if tp not in Parser._recognized_types:
                        raise SyntaxError(f"Unrecognized type '{tp}' for {name}")
            elif t not in Parser._recognized_types:
                raise SyntaxError(f"Unrecognized type '{t}' for {name}")

            # Validate that the column as a valid name.
            if not Parser.is_valid_name(name):
                raise SyntaxError(f"column '{name}' is invalid")
            # If column has a max value, validate that it is an int or float.
            if 'max' in col_content:
                try:
                    float(col_content['max'])
                except ValueError:
                    raise SyntaxError(f"{name} max value is not a number")
            # If column has a min value, validate that it is an int or float.
            if 'min' in col_content:
                try:
                    float(col_content['min'])
                except ValueError:
                    raise SyntaxError(f"{name} min value is not a number")

        generated_columns = []
        if 'generated_columns' in content:
            for generator, column in content['generated_columns'].items():
                # Validate that the generator is recognized.
                if generator not in Generators().names:
                    raise SyntaxError(f"Unrecognized generator '{generator}'")

                # Argument columns may be separated by a comma
                # to indicate a multivariate generator.
                if ',' in column:
                    columns = "".join(column.split()).split(',')
                    # Validate that all argument columns were declared.
                    for a in columns:
                        if a not in content['columns'].keys():
                            raise SyntaxError(f"Undeclared column '{a}'")
                else:
                    # Validate that the argument column was declared.
                    if column not in content['columns'].keys():
                        raise SyntaxError(f"Undeclared column '{column}'")

            generated_columns = list(content['generated_columns'])

            # Throw warning when h3 is defined but not h9.
            if ('h3' in generated_columns
                    and 'h9' not in generated_columns):
                warnings.warn("Without defining a h9 identifier, "
                              "no high accuracy queries can be made.",
                              RuntimeWarning, stacklevel=3)

        all_columns = (list(content['columns'].keys()) + generated_columns)

        # Validate if there are duplicate columns.
        for column in all_columns:
            if all_columns.count(column) > 1:
                raise SyntaxError(f"Query column {column} is duplicated")

        # Validate that backup is not empty.
        if content['backup'] is None:
            raise SyntaxError("Backup not defined.")
        # Validate that backup is recognized.
        if content['backup'] not in Parser._recognized_splits:
            raise SyntaxError(f"Unrecognized backup '{content['backup']}'")

        if 'split' in content:
            # Validate that split is not empty.
            if content['split'] is None:
                raise SyntaxError("Split not defined.")
            # Validate that split is recognized.
            if content['split'] not in Parser._recognized_splits:
                raise SyntaxError(f"Unrecognized split '{content['split']}'")
            # Validate that split is a generated column.
            # This enforces consistency in the table names.
            if ('generated_columns' not in content
                    or content['split'] not in content['generated_columns']):
                raise SyntaxError(f"Split needs to be a generated column")

        if 'lifetime' in content:
            # Validate that lifetime is not empty.
            if content['lifetime'] is None:
                raise SyntaxError("Lifetime not defined.")
            # Validate that lifetime is an integer.
            try:
                v = int(content['lifetime'])
            except ValueError:
                raise SyntaxError(
                    f"Lifetime '{content['lifetime']}' is not an integer")
            # Validate that lifetime is positive.
            if v < 0:
                raise SyntaxError(f"Lifetime '{content['lifetime']}' is not positive")

        # Validate queries is not empty.
        if content['query'] is None:
            raise SyntaxError("Query not defined.")
        # Validate that base query is declared.
        if 'base' not in content['query']:
            raise SyntaxError(f"Query 'base' is required")
        # Validate that base query is not empty.
        if content['query']['base'] is None:
            raise SyntaxError("Base query not defined.")

        base = content['query']['base']
        # Validate that base query as all necessary fields.
        missing_fields = [f for f in Parser._required_query_fields if f not in base]
        if len(missing_fields) > 0:
            missing_fields_str = ', '.join(missing_fields)
            raise SyntaxError(
                f"Base query missing required fields: {missing_fields_str}")
        # Validate that base query required field is not empty.
        if base['required'] is None:
            raise SyntaxError("Base query required columns not defined.")

        base_keys = list(base['required'].values())
        if 'optional' in base:
            base_keys += base['optional']

        for name, query in content['query'].items():
            # Validate that query has valid name.
            if not Parser.is_valid_name(name):
                raise SyntaxError(f"Query name '{name}' is invalid")

            # Validate that query has all required fields.
            missing_fields = [f for f in Parser._required_query_fields if f not in query]
            if len(missing_fields) > 0:
                missing_fields_str = ', '.join(missing_fields)
                raise SyntaxError(
                    f"{name} query missing required fields: "
                    f"{missing_fields_str}")
            # Validate that query required field is not empty.
            if query['required'] is None:
                raise SyntaxError(f"{name} query required columns "
                                  f"not defined.")
            # Validate that, if present, the optional field is not empty.
            if 'optional' in query and query['optional'] is None:
                raise \
                    SyntaxError(f"{name} query optional columns not defined")

            all_fields = Parser._required_query_fields + Parser._optional_query_fields

            # Validate that there are no unrecognized fields in query.
            unrecognized_fields = [f for f in query if f not in all_fields]
            if len(unrecognized_fields) > 0:
                unrecognized_fields_str = ', '.join(unrecognized_fields)
                raise SyntaxError(f"{name} query unrecognized fields: "
                                  f"{unrecognized_fields_str}")

            partition_columns = []
            for field, column in query['required'].items():
                # Validate that query required does not match split partition.
                # If it matched then the required field would be redundant.
                if 'split' in content and column == content['split']:
                    raise SyntaxError(
                        f"Query required field can not match split")
                # Validate that query required columns are declared as
                # the base query required or optional columns.
                # This is required by Cassandra.
                if column not in base_keys:
                    raise SyntaxError(f"{name} query required column "
                                      f"'{column}' must be a "
                                      "'base' query key, required or optional")
                # Validate that query required type is recognized.
                if field not in Parser._recognized_query_required:
                    raise SyntaxError(
                        f"{name} query unrecognized required field '{field}'")
                # Validate that query required column is declared.
                if column not in all_columns:
                    raise SyntaxError(
                        f"{name} query unrecognized required column "
                        f"'{column}'")

                partition_columns.append(column)

            if 'optional' in query:
                # Validate that query optional field is not empty.
                if query['optional'] is None:
                    raise SyntaxError(f"{name} query optional "
                                      f"columns not defined")

                for column in query['optional']:
                    # Validate that query optional column is declared.
                    if column not in all_columns:
                        raise SyntaxError(
                            f"{name} query undeclared required column "
                            f"'{column}'"
                        )

                    partition_columns.append(column)

            if 'order' in query:
                # Validate that order is not empty.
                if query['order'] is None:
                    raise SyntaxError(f"{name} query columns order "
                                      "not defined")

                for column, order in query['order'].items():
                    # Validate that order is a clustering key in query.
                    if column not in partition_columns:
                        raise SyntaxError(f"Order column {column} "
                                          "must be a clustering key")
                    # Validate that order has a recognized value.
                    if order not in Parser._recognized_orders:
                        raise SyntaxError(f"Order value {order} is not valid, "
                                          f"use either 'asc' or 'desc'")



import logging
from typing import List, Dict, Callable

from cassandra import concurrent, ConsistencyLevel
from cassandra.cluster import \
    Cluster, Session, ExecutionProfile, EXEC_PROFILE_DEFAULT
from cassandra.query import dict_factory
from cassandra.auth import AuthProvider
from cassandra.policies import \
    LoadBalancingPolicy, RetryPolicy, RoundRobinPolicy, \
    AddressTranslator


class CassandraManager:
    """Cassandra Manager class.
    This class is responsible for creating and managing the cluster connection.

    """

    @property
    def contact_points(self) -> List[str]:
        return self._contact_points

    @property
    def cluster(self) -> Cluster:
        return self._cluster

    @property
    def session(self) -> Session:
        return self._session

    @property
    def execution_profiles(self) -> Dict[str, ExecutionProfile]:
        return self._execution_profiles

    @property
    def address_translator(self) -> AddressTranslator:
        return self._address_translator

    @staticmethod
    def create_execution_profile(
        load_balancing_policy: LoadBalancingPolicy,
        retry_policy: RetryPolicy,
        consistency_level: int = ConsistencyLevel.LOCAL_ONE,
        row_factory: Callable = dict_factory,
        request_timeout: float = 10.0
    ) -> ExecutionProfile:
        """Creates a Cassandra ExecutionProfile.

        :param load_balancing_policy: load balancing policy
        :param retry_policy: retry policy used when not set on execute
        :param consistency_level: consistency level used when not set on execute
        :param row_factory: row factory (default: dict_factory)
        :param request_timeout: request timeout (default: 10.0)
        :return: Cassandra execution profile
        """

        return ExecutionProfile(
            load_balancing_policy=load_balancing_policy,
            retry_policy=retry_policy,
            consistency_level=consistency_level,
            row_factory=row_factory,
            request_timeout=request_timeout
        )

    def __init__(
        self,
        contact_points: List[str],
        connect_timeout: float = 5.0,
        control_connection_timeout: float = 2.0,
        profiles: Dict[str, ExecutionProfile] = None,
        address_translator: AddressTranslator = None,
        auth_provider: AuthProvider = None
    ):
        """Cassandra Manager constructor.

        If profiles is not defined defaults to ExecutionProfile with:
          - cassandra.policies.RoundRobinPolicy load balancing policy
          - cassandra.policies.RetryPolicy retry policy
          - LOCAL_ONE consistency level
          - dict_factory row factory
          - 10.0 seconds request_timeout

        :param contact_points: list of Cassandra node ip addresses
        :type contact_points: list[str]
        :param control_connection_timeout: connect timeout (default: 5.0)
        :type control_connection_timeout: int
        :param profiles: execution profiles (defaults:
        :param address_translator: translator to be used in translating
            server node addresses to driver connection addresses (default: None)
        :param auth_provider: authentication provider (default: None)
        :type profiles: dict
        """
        self._contact_points = contact_points
        self._address_translator = address_translator
        self._session = None

        if profiles is None:
            self._execution_profiles = {
                EXEC_PROFILE_DEFAULT: self.create_execution_profile(
                    load_balancing_policy=RoundRobinPolicy(),
                    retry_policy=RetryPolicy()
                )
            }
        else:
            self._execution_profiles = profiles

        self._cluster = Cluster(
            contact_points=self.contact_points,
            connect_timeout=connect_timeout,
            control_connection_timeout=control_connection_timeout,
            execution_profiles=self.execution_profiles,
            address_translator=self.address_translator,
            auth_provider=auth_provider
        )

    def connect(self, keyspace: str = None):
        """Connects to the Cassandra cluster, creating a session.

        :param keyspace: keyspace name
        :return: self
        """
        self._session = self.cluster.connect(keyspace=keyspace)

        return self

    def execute(
        self,
        statements: List[str],
        execution_profile: str or ExecutionProfile = None
    ) -> List[tuple] or List[dict]:
        """Execute list of query statements sequentially.

        :param statements: list of query statements
        :param execution_profile: execution profile (default: None)
            This parameter can be both the name of a configured profile,
            or the execution profile itself.
        :return: list of rows as formatted by the rows_factory
            in the execution profile
        """
        result_list = []
        for statement in statements:
            if execution_profile is not None:
                result = self.session.execute(statement, execution_profile=execution_profile)
            else:
                result = self.session.execute(statement)

            result_list += [row for row in result]

        return result_list

    def execute_concurrent(
        self,
        statements: List[str],
        raise_on_first_error: bool = False
    ) -> List[tuple] or List[dict]:
        """Execute list of query statements concurrently.

        This method does not allow to specify the execution profile to use,
        it is thus encouraged to set the desired execution profile as default
        when creating the CassandraManager object. You can do this with the
        :func:`~manager.CassandraManager.create_default_execution_profile`
        method.

        :param statements: list of query statements
        :param raise_on_first_error: raise exception on first error
            or continue and log possible errors (default: True)
        :return: list of rows as formatted by the rows_factory
            in the execution profile
        """
        result_list = []

        statements_and_params = [(s, ()) for s in statements]
        query_results = concurrent.execute_concurrent(
            self.session,
            statements_and_params,
            raise_on_first_error=raise_on_first_error
        )

        for (success, result) in query_results:
            if not success:
                logging.error(f"A query failed with error: {result}")
            else:
                result_list += [row for row in result]

        return result_list

    def close(self) -> None:
        """Close Cassandra cluster connection."""
        self.cluster.shutdown()



from datetime import datetime, timedelta

import pytz
import h3.api.basic_str as h3


class Generators:

    @property
    def names(self):
        return self._names

    def __init__(self):
        self._names = []
        for x, y in Generators.__dict__.items():
            if isinstance(y, staticmethod):
                self.names.append(x)

    @staticmethod
    def minute(ts):
        """Generate new timestamp for that minute at second zero, in UTC.

        :param ts: timestamp
        :type ts: int
        :return: datetime object for that minute at second
        """
        ts = datetime.fromtimestamp(ts / 1000, tz=pytz.UTC) \
            .replace(second=0, microsecond=0, tzinfo=pytz.UTC) \
            .timestamp()

        return int(ts * 1000)

    @staticmethod
    def hour(ts):
        """Generate new timestamp for that hour at minute zero, in UTC.

        :param ts: timestamp
        :type ts: int
        :return: datetime object for that hour at minute
        """
        ts = datetime.fromtimestamp(ts / 1000, tz=pytz.UTC) \
            .replace(minute=0, second=0, microsecond=0, tzinfo=pytz.UTC) \
            .timestamp()

        return int(ts * 1000)

    @staticmethod
    def day(ts):
        """Generate new timestamp for that day at midnight, in UTC.

        :param ts: timestamp
        :type ts: int
        :return: datetime object for that day at midnight
        """
        ts = datetime.fromtimestamp(ts / 1000, tz=pytz.UTC) \
            .replace(hour=0, minute=0, second=0, microsecond=0,
                     tzinfo=pytz.UTC).timestamp()

        return int(ts * 1000)

    @staticmethod
    def week(ts):
        """Generate new timestamp for that week on monday at midnight, in UTC.

        :param ts: timestamp
        :type ts: int
        :return: datetime object for that week on monday at midnight
        """
        d = datetime.fromtimestamp(ts / 1000, tz=pytz.UTC)
        last_monday = d - timedelta(days=d.weekday())

        ts = last_monday.replace(hour=0, minute=0, second=0, microsecond=0,
                                 tzinfo=pytz.UTC).timestamp()

        return int(ts * 1000)

    @staticmethod
    def month(ts):
        """Generate new timestamp for that month on the first day at midnight, in UTC.

        :param ts: timestamp
        :type ts: int
        :return: datetime object for that month on the first day at midnight
        """
        ts = datetime.fromtimestamp(ts / 1000, tz=pytz.UTC) \
            .replace(day=1, hour=0, minute=0, second=0, microsecond=0,
                     tzinfo=pytz.UTC).timestamp()

        return int(ts * 1000)

    @staticmethod
    def year(ts):
        """Generate new timestamp for that year on the first day at midnight, in UTC.

        :param ts: timestamp
        :type ts: int
        :return: datetime object for that year on the first day at midnight
        """
        ts = datetime.fromtimestamp(ts / 1000, tz=pytz.UTC) \
            .replace(month=1, day=1, hour=0, minute=0,
                     second=0, microsecond=0, tzinfo=pytz.UTC) \
            .timestamp()

        return int(ts * 1000)

    @staticmethod
    def h3(lat, lon):
        """Returns h3 identifier corresponding to the given identifier.

        :param lat: latitude
        :type lat: float
        :param lon: longitude
        :type lon: float
        :return: datetime object for that year on the first day at midnight
        """
        return h3.geo_to_h3(lat, lon, 3)

    @staticmethod
    def h3_begin(lat, lon):
        """Returns h3 identifier for the trip start point.
        Calls :func:`~generators.Generators.h3`"""
        return Generators.h3(lat, lon)

    @staticmethod
    def h3_end(lat, lon):
        """Returns h3 identifier for the trip end point.
        Calls :func:`~generators.Generators.h3`"""
        return Generators.h3(lat, lon)

    @staticmethod
    def h4(lat, lon):
        """Returns h4 identifier corresponding to the given identifier.

        :param lat: latitude
        :type lat: float
        :param lon: longitude
        :type lon: float
        :return: datetime object for that year on the first day at midnight
        """
        return h3.geo_to_h3(lat, lon, 4)

    @staticmethod
    def h4_begin(lat, lon):
        """Returns h4 identifier for the trip start point.
        Calls :func:`~generators.Generators.h4`"""
        return Generators.h4(lat, lon)

    @staticmethod
    def h4_end(lat, lon):
        """Returns h4 identifier for the trip end point.
        Calls :func:`~generators.Generators.h4`"""
        return Generators.h4(lat, lon)

    @staticmethod
    def h5(lat, lon):
        """Returns h5 identifier corresponding to the given identifier.

        :param lat: latitude
        :type lat: float
        :param lon: longitude
        :type lon: float
        :return: datetime object for that year on the first day at midnight
        """
        return h3.geo_to_h3(lat, lon, 5)

    @staticmethod
    def h5_begin(lat, lon):
        """Returns h5 identifier for the trip start point.
        Calls :func:`~generators.Generators.h5`"""
        return Generators.h5(lat, lon)

    @staticmethod
    def h5_end(lat, lon):
        """Returns h5 identifier for the trip end point.
        Calls :func:`~generators.Generators.h5`"""
        return Generators.h5(lat, lon)

    @staticmethod
    def h6(lat, lon):
        """Returns h6 identifier corresponding to the given identifier.

        :param lat: latitude
        :type lat: float
        :param lon: longitude
        :type lon: float
        :return: datetime object for that year on the first day at midnight
        """
        return h3.geo_to_h3(lat, lon, 6)

    @staticmethod
    def h6_begin(lat, lon):
        """Returns h6 identifier for the trip start point.
        Calls :func:`~generators.Generators.h6`"""
        return Generators.h6(lat, lon)

    @staticmethod
    def h6_end(lat, lon):
        """Returns h6 identifier for the trip end point.
        Calls :func:`~generators.Generators.h6`"""
        return Generators.h6(lat, lon)

    @staticmethod
    def h7(lat, lon):
        """Returns h7 identifier corresponding to the given identifier.

        :param lat: latitude
        :type lat: float
        :param lon: longitude
        :type lon: float
        :return: datetime object for that year on the first day at midnight
        """
        return h3.geo_to_h3(lat, lon, 7)

    @staticmethod
    def h7_begin(lat, lon):
        """Returns h7 identifier for the trip start point.
        Calls :func:`~generators.Generators.h7`"""
        return Generators.h7(lat, lon)

    @staticmethod
    def h7_end(lat, lon):
        """Returns h7 identifier for the trip end point.
        Calls :func:`~generators.Generators.h7`"""
        return Generators.h7(lat, lon)

    @staticmethod
    def h8(lat, lon):
        """Returns h8 identifier corresponding to the given identifier.

        :param lat: latitude
        :type lat: float
        :param lon: longitude
        :type lon: float
        :return: datetime object for that year on the first day at midnight
        """
        return h3.geo_to_h3(lat, lon, 8)

    @staticmethod
    def h8_begin(lat, lon):
        """Returns h8 identifier for the trip start point.
        Calls :func:`~generators.Generators.h8`"""
        return Generators.h8(lat, lon)

    @staticmethod
    def h8_end(lat, lon):
        """Returns h8 identifier for the trip end point.
        Calls :func:`~generators.Generators.h8`"""
        return Generators.h8(lat, lon)

    @staticmethod
    def h9(lat, lon):
        """Returns h9 identifier corresponding to the given identifier.

        :param lat: latitude
        :type lat: float
        :param lon: longitude
        :type lon: float
        :return: datetime object for that year on the first day at midnight
        """
        return h3.geo_to_h3(lat, lon, 9)

    @staticmethod
    def h9_begin(lat, lon):
        """Returns h9 identifier for the trip start point.
        Calls :func:`~generators.Generators.h9`"""
        return Generators.h9(lat, lon)

    @staticmethod
    def h9_end(lat, lon):
        """Returns h9 identifier for the trip end point.
        Calls :func:`~generators.Generators.h9`"""
        return Generators.h9(lat, lon)

    @staticmethod
    def h10(lat, lon):
        """Returns h10 identifier corresponding to the given identifier.

        :param lat: latitude
        :type lat: float
        :param lon: longitude
        :type lon: float
        :return: datetime object for that year on the first day at midnight
        """
        return h3.geo_to_h3(lat, lon, 10)

    @staticmethod
    def h10_begin(lat, lon):
        """Returns h10 identifier for the trip start point.
        Calls :func:`~generators.Generators.h10`"""
        return Generators.h10(lat, lon)

    @staticmethod
    def h10_end(lat, lon):
        """Returns h10 identifier for the trip end point.
        Calls :func:`~generators.Generators.h10`"""
        return Generators.h10(lat, lon)

    @staticmethod
    def h11(lat, lon):
        """Returns h11 identifier corresponding to the given identifier.

        :param lat: latitude
        :type lat: float
        :param lon: longitude
        :type lon: float
        :return: datetime object for that year on the first day at midnight
        """
        return h3.geo_to_h3(lat, lon, 11)

    @staticmethod
    def h11_begin(lat, lon):
        """Returns h11 identifier for the trip start point.
        Calls :func:`~generators.Generators.h11`"""
        return Generators.h11(lat, lon)

    @staticmethod
    def h11_end(lat, lon):
        """Returns h11 identifier for the trip end point.
        Calls :func:`~generators.Generators.h11`"""
        return Generators.h11(lat, lon)

    @staticmethod
    def h12(lat, lon):
        """Returns h12 identifier corresponding to the given identifier.

        :param lat: latitude
        :type lat: float
        :param lon: longitude
        :type lon: float
        :return: datetime object for that year on the first day at midnight
        """
        return h3.geo_to_h3(lat, lon, 12)

    @staticmethod
    def h12_begin(lat, lon):
        """Returns h12 identifier for the trip start point.
        Calls :func:`~generators.Generators.h12`"""
        return Generators.h12(lat, lon)

    @staticmethod
    def h12_end(lat, lon):
        """Returns h12 identifier for the trip end point.
        Calls :func:`~generators.Generators.h12`"""
        return Generators.h12(lat, lon)




class NoConnectionError(Exception):
    """Exception thrown when the Cassandra Cluster cannot be reached."""

    def __init__(self, seed):
        super(NoConnectionError, self) \
            .__init__(f"The Cassandra cluster '{seed}' can not be reached.")


class DateNotDefinedError(Exception):
    """Exception thrown when date is required but not defined."""

    def __init__(self, message):
        super(DateNotDefinedError, self).__init__(message)


class MissingColumnError(Exception):
    """Exception thrown when a column is required but not defined."""

    def __init__(self, message):
        super(MissingColumnError, self).__init__(message)


class QueryNotFound(Exception):
    """Exception thrown when the specified query name is not defined."""

    def __init__(self, name):
        super(QueryNotFound, self).__init__(f"Query '{name}' not found.")


class NotARequiredColumnError(Exception):
    """Exception thrown when a query specifies a required column that is not defined in the yaml."""

    def __init__(self, name, query):
        super(NotARequiredColumnError, self) \
            .__init__(f"'{name}' is not a required column in query '{query}'.")



from pathlib import Path
import logging

import yaml

from primeight.parser.parser import Parser


logger = logging.getLogger(__name__)


class YamlParser(Parser):

    @staticmethod
    def parse(path: str or Path) -> dict:
        """Load Yaml from disk."""
        config_file = path if isinstance(path, Path) else Path(path)

        if config_file.is_dir():
            raise IsADirectoryError(f"Table yaml '{path}' is a directory.")
        if not config_file.is_file():
            raise FileNotFoundError(f"Table yaml '{path}' not found.")

        yaml_config = config_file.read_text(encoding='utf-8')
        if len(yaml_config) == 0:
            raise EOFError("Yaml file is empty")

        try:
            content = yaml.safe_load(yaml_config)
        except yaml.MarkedYAMLError as err:
            mark = err.problem_mark
            raise SyntaxError(
                f"Table yaml '{path}' is not a valid yaml. "
                f"Error position: {mark['line']}:{mark['column']}"
            )
        except yaml.YAMLError:
            raise SyntaxError(f"Table yaml '{path}' is not a valid yaml.")

        YamlParser.is_valid_config(content)
        return content



import uuid
import logging
from typing import List, Tuple, Set


logger = logging.getLogger(__name__)


class CassandraColumn:

    NATIVE_TYPES = [
        'ascii', 'bigint', 'blob', 'boolean', 'counter', 'decimal', 'double',
        'float', 'inet', 'int', 'smallint', 'tinyint', 'text', 'date', 'time',
        'timestamp', 'timeuuid', 'duration', 'uuid', 'varchar', 'varint', 'h3hex'
    ]
    COLLECTIONS = ['list', 'set', 'map']
    OTHER_TYPES = ['tuple']

    HANDLE_TO_TYPE = {
        'ascii': str,
        'bigint': int,
        'blob': bytes,
        'boolean': bool,
        'counter': int,
        'decimal': float,
        'double': float,
        'float': float,
        'inet': str,
        'int': int,
        'smallint': int,
        'tinyint': int,
        'text': str,
        'date': str,
        'time': int,
        'timestamp': int,
        'timeuuid': uuid.UUID,
        'duration': str,
        'uuid': uuid.UUID,
        'varchar': str,
        'varint': int,
        'h3hex': str
    }

    HANDLE_TO_CASSANDRA_TYPE = {
        'ascii': 'ASCII',
        'bigint': 'BIGINT',
        'blob': 'BLOB',
        'boolean': 'BOOLEAN',
        'counter': 'COUNTER',
        'decimal': 'DECIMAL',
        'double': 'DOUBLE',
        'float': 'FLOAT',
        'inet': 'INET',
        'int': 'INT',
        'smallint': 'SMALLINT',
        'tinyint': 'TINYINT',
        'text': 'TEXT',
        'date': 'DATE',
        'time': 'TIME',
        'timestamp': 'BIGINT',
        'timeuuid': 'TIMEUUID',
        'duration': 'DURATION',
        'uuid': 'UUID',
        'varchar': 'VARCHAR',
        'varint': 'VARINT',
        'h3hex': 'TEXT'
    }

    @property
    def name(self):
        return self._name

    @property
    def type(self):
        return self._type_handle

    @property
    def min_value(self):
        if self._min is None:
            logger.debug("Column %s has no min value defined.", self.name)

        return self._min

    @property
    def max_value(self):
        if self._max is None:
            logger.debug("Column %s has no max value defined.", self.name)

        return self._max

    @property
    def alias(self):
        if self._alias is None:
            logger.debug("Column %s has no alias defined.", self.name)

        return self._alias

    @property
    def description(self):
        if self._description is None:
            logger.debug("Column %s has no alias defined.", self.name)

        return self._description

    def __init__(
        self,
        name: str,
        type_handle: str,
        alias: str = None,
        description: str = None,
        min_value: int or float = None,
        max_value: int or float = None
    ):
        """Cassandra column constructor.

        :param name: column name
        :param type_handle: column type handle
        :param alias: column alias to the name (default: None)
        :param description: column description (default: None)
        :param min_value: column minimum value.
            This is used to check for invalid data (default: None)
        :param max_value: column maximum value.
            This is used to check for invalid data (default: None)
        """
        self._name = name.lower()
        self._type_handle = type_handle.lower()

        self._min = min_value
        self._max = max_value
        self._alias = alias
        self._description = description

    def pydantic_type(self, handle=None):
        if handle is None:
            handle = self._type_handle

        if handle.startswith('list'):
            h = handle \
                .replace('list<', '', 1) \
                .replace('>', '')
            content = self.pydantic_type(handle=h)

            return List[content]
        elif handle.startswith('tuple'):
            handle = handle \
                .replace('tuple<', '', 1) \
                .replace('>', '') \
                .replace(' ', '')
            content = tuple(self.HANDLE_TO_TYPE[h] for h in handle.split(','))

            return Tuple[content]
        elif handle.startswith('set'):
            h = handle \
                .replace('set<', '', 1) \
                .replace('>', '')
            content = self.pydantic_type(handle=h)

            return Set[content]
        else:
            return self.HANDLE_TO_TYPE[handle]

    @staticmethod
    def _handle_cassandra_frozen(handle):
        return f"FROZEN<{handle}>"

    def _handle_cassandra_list_type(self, handle, frozen=True):
        handle = handle \
            .lower() \
            .replace(' ', '') \
            .replace('list<', '', 1) \
            .strip('>')

        contents = self.cassandra_type(handle, frozen=frozen)
        type_handle = f'LIST<{contents}>'
        if frozen:
            type_handle = CassandraColumn._handle_cassandra_frozen(type_handle)

        return type_handle

    def _handle_cassandra_set_type(self, handle, frozen=True):
        handle = handle \
            .lower() \
            .replace(' ', '') \
            .replace('set<', '', 1) \
            .strip('>')

        type_handle = f'SET<{self.cassandra_type(handle)}>'
        if frozen:
            type_handle = self._handle_cassandra_frozen(type_handle)

        return type_handle

    def _handle_cassandra_tuple_type(self, handle, frozen=True):
        handle = handle \
            .lower() \
            .replace(' ', '') \
            .replace('tuple<', '', 1) \
            .strip('>')

        content = ','.join([self.cassandra_type(t) for t in handle.split(',')])
        type_handle = f'TUPLE<{content}>'
        if frozen:
            type_handle = self._handle_cassandra_frozen(type_handle)

        return type_handle

    def _handle_cassandra_map_type(self, handle, frozen=True):
        handle = handle \
            .lower() \
            .replace(' ', '') \
            .replace('map<', '', 1) \
            .strip('>')

        content = ','.join([self.cassandra_type(t) for t in handle.split(',')])
        type_handle = f'MAP<{content}>'
        if frozen:
            type_handle = self._handle_cassandra_frozen(type_handle)

        return type_handle

    def cassandra_type(self, handle: str = None, frozen: bool = True) -> str:
        type_handle = handle or self._type_handle

        if type_handle.startswith('list'):
            return self._handle_cassandra_list_type(type_handle, frozen=frozen)
        elif type_handle.startswith('set'):
            return self._handle_cassandra_set_type(type_handle, frozen=frozen)
        elif type_handle.startswith('map'):
            return self._handle_cassandra_map_type(type_handle, frozen=frozen)
        elif type_handle.startswith('tuple'):
            return self._handle_cassandra_tuple_type(type_handle, frozen=frozen)
        else:
            return self.HANDLE_TO_CASSANDRA_TYPE[type_handle]

    def is_valid(self, value):
        """Returns True if value is valid, and False otherwise."""

        if self.max_value is not None and value > self.max_value:
            logger.debug("value '%s' is above '%s'", value, self.max_value)
            return False

        if self.min_value is not None and value < self.min_value:
            logger.debug("value '%s' is below '%s'", value, self.min_value)
            return False

        return True



from typing import List

from cassandra.cluster import ExecutionProfile

from primeight import CassandraManager


class CassandraBase:

    @property
    def cassandra_manager(self) -> CassandraManager:
        """Returns the Cassandra manager."""
        return self._cassandra_manager

    @property
    def config(self) -> dict:
        """Returns the table config."""
        return self._config

    def __init__(
        self, config: dict, cassandra_manager: CassandraManager = None
    ):
        """Cassandra base constructor.

        :param config: table template configuration
        :param cassandra_manager: cassandra manager
        """
        self._config = config
        self._cassandra_manager = cassandra_manager

        self._current_statements = None

    def execute(
            self, execution_profile: str or ExecutionProfile = None
    ) -> List[tuple] or List[dict]:
        """Execute list of query statements sequentially.

        :param execution_profile: execution profile (default: None)
            This parameter can be both the name of a configured profile,
            or the execution profile itself.
        :return: list of rows as formatted by the rows_factory
            in the execution profile
        """
        result = self.cassandra_manager.execute(self.statements, execution_profile)

        return result

    def execute_concurrent(
        self,
        raise_on_first_error: bool = False
    ) -> List[tuple] or List[dict]:
        """Execute list of query statements concurrently.

        This method does not allow to specify the execution profile to use,
        it is thus encouraged to set the desired execution profile as default
        when creating the CassandraManager object. You can do this with the
        :func:`~manager.CassandraManager.create_default_execution_profile`
        method.

        :param raise_on_first_error: raise exception on first error
            or continue and log possible errors (default: True)
        :return: list of rows as formatted by the rows_factory
            in the execution profile
        """
        result = self.cassandra_manager.execute_concurrent(
            self.statements, raise_on_first_error
        )

        return result



from typing import List, Dict

from cassandra.query import Statement, SimpleStatement

from primeight.base import CassandraBase
from primeight.manager import CassandraManager


class CassandraKeyspace(CassandraBase):
    """Cassandra keyspace class.
    This class incorporates all the information and operations
    allowed over keyspaces.

    """

    @property
    def config(self) -> dict:
        """Returns template configuration."""
        return self._config

    @property
    def statements(self) -> List[str] or List[Statement]:
        """Returns Cassandra statements."""
        if self._current_statements is None:
            return []

        return self._current_statements

    @property
    def name(self) -> str:
        """Returns keyspace name."""
        if 'keyspace' not in self.config:
            raise ValueError("Missing keyspace name.")
        return self.config['keyspace']

    def __init__(self, config: dict, cassandra_manager: CassandraManager = None):
        """Cassandra keyspace constructor.

        :param config: table template configuration
        :param cassandra_manager: cassandra manager
        """
        super().__init__(config, cassandra_manager)

        self._config = config
        self._current_statements = []

    def create(
        self,
        name: str or List[str] = None,
        replication_strategy='SimpleStrategy',
        replication_factor: int or Dict[str, int] = 3,
        if_not_exists=False
    ):
        """Create new keyspace statement.

        When using the `SimpleStrategy` replication strategy,
        `replication_factor` should be an integer.
        When using the `NetworkTopologyStrategy` replication strategy,
        `replication_factor` should be a dictionary mapping data center names
        to an integer value representing the replication factor in that
        data center.

        :param name: keyspace name.
            This may be an str or List[str]  (default: None)
        :param replication_strategy: keyspace replication strategy
            (default: SimpleStrategy)
        :param replication_factor: keyspace replication factor (default: 3)
        :param if_not_exists: if True, creates keyspace if it does not exist.
            When set to False and the keyspace exists, an exception is thrown.
            (default: False)
        :return: self
        """
        keyspaces = name
        if keyspaces is None:
            keyspaces = [self.name]
        elif isinstance(name, str):
            keyspaces = [name]

        self._current_statements = []
        for keyspace_name in keyspaces:
            statement = "CREATE KEYSPACE "
            if if_not_exists:
                statement += "IF NOT EXISTS "

            statement += (
                f"{keyspace_name} "
                "WITH replication={ "
                f"'class' : '{replication_strategy}' "
            )

            if isinstance(replication_factor, int):
                statement += f", 'replication_factor': {replication_factor} "
            elif isinstance(replication_factor, dict):
                for dc_name, repl_factor in replication_factor.items():
                    statement += f", '{dc_name}': {repl_factor} "
            else:
                raise Exception("replication_factor has wrong type")

            statement += "};"

            self._current_statements.append(SimpleStatement(statement))

        return self

    def drop(self, name: str or List[str] = None, if_exists: bool = False):
        """Drop keyspace statement.

        :param name: keyspace name.
            This may be an str or List[str]  (default: None)
        :param if_exists: if True, drops keyspace if it exists.
            When set to False and the keyspace does not exist,
            an exception is thrown. (default: False)
        :return: self
        """
        keyspaces = name
        if keyspaces is None:
            keyspaces = [self.name]
        elif isinstance(name, str):
            keyspaces = [name]

        self._current_statements = []
        for keyspace_name in keyspaces:
            statement = f"DROP KEYSPACE "
            if if_exists:
                statement += "IF EXISTS "
            statement += f"{keyspace_name} ;"

            self._current_statements.append(SimpleStatement(statement))

        return self

    def alter(
        self,
        name: str or List[str] = None,
        replication_strategy='SimpleStrategy',
        replication_factor: int or Dict[str, int] = 3
    ):
        """Alter keyspace statement.

        When using the `SimpleStrategy` replication strategy,
        `replication_factor` should be an integer.
        When using the `NetworkTopologyStrategy` replication strategy,
        `replication_factor` should be a dictionary mapping data center names
        to an integer value representing the replication factor in that
        data center.

        :param name: keyspace name.
            This may be an str or List[str]  (default: None)
        :param replication_strategy: keyspace replication strategy
            (default: SimpleStrategy)
        :param replication_factor: keyspace replication factor (default: 3)
        :return: self
        """
        keyspaces = name
        if keyspaces is None:
            keyspaces = [self.name]
        elif isinstance(name, str):
            keyspaces = [name]

        self._current_statements = []
        for keyspace_name in keyspaces:
            statement = (
                f"ALTER KEYSPACE {keyspace_name} "
                "WITH replication={ "
                f"'class' : '{replication_strategy}' "
            )

            if isinstance(replication_factor, int):
                statement += f", 'replication_factor': {replication_factor} "
            elif isinstance(replication_factor, dict):
                for dc_name, repl_factor in replication_factor.items():
                    statement += f", '{dc_name}': {repl_factor} "
            else:
                raise Exception("replication_factor has wrong type")

            statement += "};"

            self._current_statements.append(SimpleStatement(statement))

        return self



from setuptools import setup
from pathlib import Path

with Path('README.md').open('r') as f:
    long_description = f.read()

with Path('requirements.txt').open('r') as f:
    requires = f.read().splitlines()

setup(
    name='primeight',
    version='0.1.2',
    url='https://pri.meight.com',
    author='Meight',
    author_email='engineering@meight.com',
    description="An Open Source Python package to manage and interact with Cassandra",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=['primeight', 'primeight.parser'],
    install_requires=requires,
    license='Apache License 2.0',
    python_requires='>=3.7',
    classifiers=[
        "Programming Language :: Python :: 3",
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ]
)



import json
import logging
from datetime import datetime, timedelta
from typing import Any, List, Dict, Optional

import pytz
from cassandra.encoder import cql_quote
from pydantic import create_model
import h3.api.basic_str as h3
from geojson import Polygon

from primeight.base import CassandraBase
from primeight.manager import CassandraManager
from primeight.keyspace import CassandraKeyspace
from primeight.column import CassandraColumn
from primeight.generators import Generators
from primeight.utils import UUIDEncoder
from primeight.exceptions import \
    DateNotDefinedError, QueryNotFound, \
    MissingColumnError, NotARequiredColumnError


class CassandraTable(CassandraBase):

    TABLE_FORMAT = {
        'day': '%d_%m_%Y',
        'week': '%d_%m_%Y',
        'month': '%m_%Y',
        'year': '%Y'
    }

    @property
    def name(self):
        """Returns the table name."""
        return self.config['name']

    @property
    def keyspace(self) -> CassandraKeyspace:
        """Returns the table keyspace object."""
        if self._keyspace is None:
            raise ValueError("Cassandra keyspace not specified.")

        return self._keyspace

    @property
    def keyspace_name(self) -> str:
        """Returns the table keyspace name."""
        try:
            _name = self.keyspace.name
        except ValueError:
            _name = self.config['keyspace']

        return _name

    @property
    def columns(self) -> List[CassandraColumn]:
        """Returns the table column list."""
        return self._columns

    @property
    def col(self) -> Dict[str, CassandraColumn]:
        return {col.name.lower(): col for col in self.columns}

    @property
    def model(self):
        """Returns a Pydantic model of the table arguments."""
        name = self.config['name'].title().replace('_', '')
        fields = {}
        for f in self.columns:
            fields[f.name] = (f.pydantic_type(), None)

        return create_model(name, **fields)

    @property
    def statements(self) -> list:
        """Returns current statements."""
        if self._current_statements is None:
            return []

        _tags = ['where', 'and', 'limit']
        _statements = []
        for statement in self._current_statements:
            if '{columns}' in statement:
                statement = self._replace(statement, 'columns', '*')

            for tag in _tags:
                _tag = '{' + tag + '}'
                if _tag in statement:
                    statement = self._replace(statement, tag, '')

            _statements.append(statement)

        return _statements

    def get_columns(
        self, names: List[str] = None, alias: List[str] = None
    ) -> List[CassandraColumn]:
        """Returns list of columns."""
        column_list = []
        for col in self.columns:
            if names is not None and col.name in names:
                column_list.append(col)

            elif alias is not None and col.alias in alias:
                column_list.append(col)

        return column_list

    def get(self, name: str) -> Optional[CassandraColumn]:
        """Returns column, returns None if column does not exist in table."""
        return self.col.get(name.lower())

    def __init__(
            self,
            config: dict,
            keyspace: CassandraKeyspace = None,
            cassandra_manager: CassandraManager = None
    ):
        """Cassandra table constructor.

        :param config: table template configuration
        :param keyspace: table keyspace
        """
        super().__init__(config, cassandra_manager)

        self._config = config
        self._keyspace = keyspace

        columns = []
        for name, content in self._config['columns'].items():
            col_min = content.get('min', None)
            col_max = content.get('max', None)
            col_alias = content.get('alias', None)
            col_description = content.get('description', None)

            col = CassandraColumn(
                name, content['type'],
                min_value=col_min, max_value=col_max,
                alias=col_alias, description=col_description
            )
            columns.append(col)

        if 'generated_columns' in self._config:
            for name, content in self._config['generated_columns'].items():
                if name in ['day', 'week', 'month', 'year']:
                    col = CassandraColumn(name, 'timestamp')
                else:
                    col = CassandraColumn(name, 'h3hex')

                columns.append(col)

        self._columns = columns

        self._current_operation = None
        self._current_query = None
        self._current_statements = None

        self._manager = cassandra_manager

    def has_split(self) -> bool:
        """Returns True if table has split, False otherwise."""
        return 'split' in self.config

    def _is_query_valid(self) -> bool:
        """Raises an Exception if statement is not ready to be executed."""

        for statement in self._current_statements:
            if '{date}' in statement:
                raise DateNotDefinedError(
                    "When splitting table by date, "
                    "you are required to specify a time frame."
                )

        return True

    @staticmethod
    def _calculate_table_partitions(partition: str, start: datetime, end: datetime) -> List[datetime]:
        """Create list of datetime from start to end date, based on partition.
        Partition may be day, month, or year.
        All datetime are return representing the initial moment of the start
        of that partition.

        :param partition: partition
        :param start: start date
        :param end: end date
        :return: list of datetime
        """

        if partition == 'day':
            dd = (end - start).days + 1

            days = []
            for i in range(dd):
                d = start + timedelta(days=i)
                days.append(d.replace(hour=0, minute=0, second=0, microsecond=0, tzinfo=pytz.UTC))

            return days
        elif partition == 'week':
            weeks = set()

            dd = (end - start).days + 1
            for i in range(dd):
                d = start + timedelta(days=i)
                last_monday = d - timedelta(days=d.weekday())
                weeks.add(last_monday)

            return list(weeks)
        elif partition == 'month':

            def total_months(dt):
                return dt.month + 12 * dt.year

            months = []
            for tot_m in range(total_months(start) - 1, total_months(end)):
                y, m = divmod(tot_m, 12)
                months.append(datetime(y, m + 1, 1, 0, 0, 0, 0, tzinfo=pytz.UTC))

            return months
        elif partition == 'year':
            dy = (end.year - start.year) + 1

            years = []
            for i in range(dy):
                y = start.year + i
                years.append(start.replace(year=y, month=1, day=1,
                                           hour=0, minute=0, second=0,
                                           microsecond=0, tzinfo=pytz.UTC))

            return years

    @staticmethod
    def _replace(statement, column, value):
        """Format statement without throwing missing key exception.
        This is required because we want to sequentially define our statement.

        :param statement: statement
        :type statement: str
        :param column: column to replace
        :type column: str
        :param value: value to replace
        :type value: str
        :return: statement with the column replaced
        """

        class PartialFormatDict(dict):

            def __missing__(self, key):
                return '{' + key + '}'

        mapping = {column: value}
        return statement.format_map(PartialFormatDict(mapping))

    def create(
        self,
        keyspace: str or List[str] = None,
        gc_grace_seconds: int = 86400,
        if_not_exists: bool = False,
        create_materialized_views: bool = True
    ):
        """Create table statement.

        :param keyspace: keyspace name.
            This may be an str or List[str]  (default: None)
        :param gc_grace_seconds: garbage collect grace seconds.
            Defines the time in seconds that a tombstone is kept before being
            discarded.
        :param if_not_exists: if True, creates table if it does not exist.
            When set to False and the table exists, an exception is thrown.
            (default: False)
        :param create_materialized_views: if True, creates materialized views
            (default: True)
        :return: self
        """
        self._current_operation = 'create'

        keyspaces = keyspace
        if keyspaces is None:
            keyspaces = [self.keyspace_name]
        elif isinstance(keyspace, str):
            keyspaces = [keyspace]

        self._current_statements = []
        for keyspace_name in keyspaces:
            base = self._config['query']['base']

            statement = f"CREATE TABLE "

            if if_not_exists:
                statement += "IF NOT EXISTS "

            statement += f"{keyspace_name}.{self.name}"
            if self.has_split():
                statement += "_{date}"

            columns = \
                ', '.join([f'{a.name} {a.cassandra_type()}' for a in self.columns])
            primary_keys = ', '.join([pk for _, pk in base['required'].items()])
            statement += f" ( {columns}, PRIMARY KEY ( ({primary_keys})"

            if 'optional' in base:
                clustering_keys = ', '.join([ck for ck in base['optional']])
                statement += f", {clustering_keys} ) ) "
            else:
                statement += " ) ) "

            statement += "WITH"

            if 'order' in base:
                statement += f" CLUSTERING ORDER BY ( "
                for index, (column, order) in enumerate(base['order'].items()):
                    if index == 0:
                        statement += f"{column} {order.upper()}"
                    else:
                        statement += f", {column} {order.upper()}"

                statement += " ) AND"

            statement += f" gc_grace_seconds={gc_grace_seconds};"
            self._current_statements.append(statement)

        if create_materialized_views and len(self.config['query']) > 1:
            for name, query_config in self.config['query'].items():
                if name == 'base':
                    continue

                self._current_statements += \
                    CassandraMaterializedView(self.config, name, keyspace=keyspace) \
                    .create(
                        keyspace=keyspace,
                        gc_grace_seconds=gc_grace_seconds,
                        if_not_exists=if_not_exists
                    ).statements

        return self

    def drop(
        self,
        keyspace: str or List[str] = None,
        if_exists: bool = False,
        drop_materialized_views: bool = True
    ):
        """Drop table statement.

        :param keyspace: keyspace name.
            This may be an str or List[str]  (default: None)
        :param if_exists: if True, drops table if it exist.
            When set to False and the table does not exist,
            an exception is thrown (default: False)
        :param drop_materialized_views: if True, creates materialized views
            (default: True)
        :return: self
        """
        self._current_operation = 'drop'

        keyspaces = keyspace
        if keyspaces is None:
            keyspaces = [self.keyspace_name]
        elif isinstance(keyspace, str):
            keyspaces = [keyspace]

        self._current_statements = []
        if drop_materialized_views and len(self.config['query']) > 1:
            for name, query_config in self.config['query'].items():
                if name == 'base':
                    continue

                self._current_statements += \
                    CassandraMaterializedView(self.config, name, self.keyspace) \
                    .drop(keyspace=keyspace, if_exists=if_exists) \
                    .statements

        for keyspace_name in keyspaces:
            statement = f"DROP TABLE "
            if if_exists:
                statement += "IF EXISTS "

            statement += f"{keyspace_name}.{self.name}"
            if self.has_split():
                statement += "_{date}"
            statement += ";"
            self._current_statements.append(statement)

        return self

    def insert(
            self,
            row: Dict[str, Any],
            keyspace: str or List[str] = None,
            ttl: int = None
    ):
        """Insert row into Cassandra.
        This method creates an insert statement that can be chained with
        other insert statements.

        When Cassandra does an insert and the row already exists,
        it overwrites the current column values.

        Also, for Cassandra, `None` row values are considered as empty,
        meaning that if you overwrite a column with a `None` value
        it will delete that row column.
        This means that you need to be careful whenever making inserts,
        so that you do not delete any undesired columns.

        :param row: values to insert in row
        :param keyspace: keyspace name.
            This may be an str or List[str]  (default: None)
        :param ttl: time to live
        :return: self
        """
        keyspaces = keyspace
        if keyspaces is None:
            keyspaces = [self.keyspace_name]
        elif isinstance(keyspace, str):
            keyspaces = [keyspace]

        # Create generated columns using the predefined generators.
        generated_columns = {}
        if 'generated_columns' in self.config:
            for name, column in self.config['generated_columns'].items():
                generator = getattr(Generators, name)

                if ',' in column:
                    columns = "".join(column.split()).split(',')
                    values = [row[a] for a in columns]
                    generated_columns[name] = generator(*values)
                else:
                    generated_columns[name] = generator(row[column])

        json_values = json.dumps({**row, **generated_columns}, cls=UUIDEncoder)

        statements = []
        for keyspace_name in keyspaces:
            statement = f"INSERT INTO {keyspace_name}.{self.name}"
            if self.has_split():
                split = self.config['split']
                split_col = self.config['generated_columns'][split]

                date = \
                    datetime.fromtimestamp(row[split_col] / 1000, tz=pytz.UTC)
                split_date = self._calculate_table_partitions(split, date, date)[0]
                date_str = split_date.strftime(self.TABLE_FORMAT[split])

                statement += f"_{date_str}"
            statement += f" JSON \'{json_values}\' "

            if ttl is not None:
                statement += f"USING TTL {ttl}"
            statement += ";"

            statements.append(statement)

        if self._current_operation == 'insert':
            self._current_statements += statements
        else:
            self._current_operation = 'insert'
            self._current_statements = statements

        return self

    def query(self, name: str = 'base', keyspace: str or List[str] = None):
        """Query statement.
        Query must be declared in the Yaml templates file.

        If left clients is not defined, it will look in the yaml
        for the keyspace field.

        This method must be chained with the methods required to define
        the primary and clustering keys, and finalizing with
        :func:`~table.CassandraTable.execute`

        :param name: query name (default: base)
        :param keyspace: keyspace name.
            This may be an str or List[str]  (default: None)
        :return: self
        """
        if name not in self.config['query']:
            raise QueryNotFound(name)

        self._current_operation = 'query'
        self._current_query = name

        keyspaces = keyspace
        if keyspaces is None:
            keyspaces = [self.keyspace_name]
        elif isinstance(keyspace, str):
            keyspaces = [keyspace]

        self._current_statements = []
        for keyspace_name in keyspaces:
            if name == 'base':
                statement = (
                    "SELECT {columns} FROM "
                    f"{keyspace_name}.{self.name}"
                )
                if self.has_split():
                    statement += "_{date}"
                statement += " {where} {limit} ;"
                self._current_statements.append(statement)
            else:
                self._current_statements += \
                    CassandraMaterializedView(self.config, name, self.keyspace) \
                    .query(keyspace=keyspace) \
                    .statements

        return self

    def select(self, columns: List[str]):
        """Select which columns to query.
        If not called it will load all columns.

        :param columns: list of column names to retrieve
        :return: self
        """
        column_names = [a.name for a in self.columns]
        for column in columns:
            if column not in column_names:
                raise MissingColumnError(f"{column} not in table columns")

        cols_str = ", ".join(columns)
        self._current_statements = [self._replace(s, 'columns', cols_str)
                                    for s in self._current_statements]

        return self

    def _prepare_column_clause_single(self, statement: str, column_name: str):
        """Prepare statement to specify required columns.
        This method is required because we allow the query of the whole table.

        :param statement: cassandra statement
        :param column_name: column name to prepare for
        :return: list of prepared statements
        """
        if '{where}' in statement:
            prefix = "WHERE "
        else:
            prefix = "AND "
        key = '{' + column_name + '}'
        sufix = " {and} "

        return self._replace(
            statement,
            prefix.lower().replace(' ', ''),
            f"{prefix}{key}{sufix}"
        )

    def _prepare_column_clause(self, column_name: str):
        """Prepare statements to specify required columns.
        This method is required because we allow the query of the whole table.

        :param column_name: column name to prepare for
        :return: list of prepared statements
        """
        statements = []
        for statement in self._current_statements:
            if '{where}' in statement:
                prefix = "WHERE "
            else:
                prefix = "AND "
            key = '{' + column_name + '}'
            sufix = " {and}"

            statements.append(self._replace(
                statement,
                prefix.lower().replace(' ', ''),
                f"{prefix}{key}{sufix}"
            ))

        return statements

    def time(
        self,
        start: datetime, end: datetime,
        prepare: bool = False, split_only: bool = False
    ):
        """Select query time frame.

        Time must be one of the required columns in the table yaml.

        :param start: start date in UTC
        :param end: end date in UTC
        :param prepare: if True, will only define table name
            and (if applicable) will leave the required key as
            a named parameter (default: False)
        :param split_only: if True, will only replace the table split
            (default: False)
        :return: self
        """

        if self.has_split():
            split = self.config['split']
            split_date_list = self._calculate_table_partitions(split, start, end)

            statements = []
            for statement in self._current_statements:
                if '{date}' in statement:
                    for date in split_date_list:
                        date_str = date.strftime(self.TABLE_FORMAT[split])
                        statements.append(
                            self._replace(statement, 'date', date_str))
                else:
                    statements.append(statement)

            self._current_statements = statements

        if not split_only and self._current_operation == 'query':
            query = self.config['query'][self._current_query]

            if 'time' not in query['required'] and not self.has_split():
                raise NotARequiredColumnError('time', self._current_query)
            elif 'time' in query['required']:
                partition = query['required']['time']

                if self.has_split() and self.config['split'] == partition:
                    return self

                self._current_statements = \
                    self._prepare_column_clause(partition)

                date_list = self._calculate_table_partitions(partition, start, end)

                statements = []
                for statement in self._current_statements:
                    if prepare:
                        replacement = f"{partition}=?"
                        s = self._replace(statement, partition, replacement)
                        statements.append(s)
                    else:
                        for date in date_list:
                            if self.has_split():
                                split = self.config['split']
                                date_str = \
                                    date.strftime(self.TABLE_FORMAT[split])
                                if f'{self.name}_{date_str}' in statement:
                                    ts = int(date.timestamp() * 1000)
                                    replacement = f"{partition}={ts}"
                                    s = self._replace(
                                        statement, partition, replacement
                                    )

                                    statements.append(s)
                            else:
                                ts = int(date.timestamp() * 1000)
                                replacement = f"{partition}={ts}"
                                s = self._replace(
                                    statement, partition, replacement
                                )

                                statements.append(s)
                self._current_statements = statements

        return self

    def space(self, identifier: str or List[str] = None):
        """Select query spacial region.

        Space must be one of the required columns in the table yaml.

        If `identifier` parameter is missing or set to `None`,
        it will be set as a named parameter.
        This method accepts either a list of identifiers or a single value.
        If it receives a list it will
        behave like :func:`~table.CassandraTable.among`.
        If it receives a value it will
        behave like :func:`~table.CassandraTable.equals`.

        :param identifier: H3 identifier or list of H3 identifiers
        :return: self
        """

        def _log_geojson(_identifiers):
            features = []
            for _tmp in _identifiers:
                hex_json = {
                    'type': 'Feature',
                    'geometry':
                        Polygon([h3.h3_to_geo_boundary(_tmp, geo_json=True)]),
                    'properties': {}
                }
                features.append(hex_json)

            geojson_dumps = json.dumps({'type': 'FeatureCollection',
                                        'features': features})
            logging.debug(geojson_dumps)

        query = self.config['query'][self._current_query]
        if 'space' not in query['required']:
            raise NotARequiredColumnError('space', self._current_query)

        level = query['required']['space']
        if identifier is None:
            replacement = f"{level}='?'"
        elif isinstance(identifier, list):
            ids_str = ', '.join([f"'{i}'" for i in identifier])
            replacement = f"{level} IN ({ids_str})"
        else:
            replacement = f"{level}='{identifier}'"

        statements = self._prepare_column_clause(level)
        self._current_statements = \
            [self._replace(s, level, replacement) for s in statements]

        logger = logging.getLogger()
        if logger.level == logging.DEBUG:
            # Since creating a geojson takes a lot of time,
            # we only log the geojson when in logging level DEBUG.
            _log_geojson([_h3 for _h3 in identifier if h3.h3_is_valid(_h3)])

        return self

    def id(self, identifier: Any or List[Any] = None):
        """Select list of ids to query.

        Id must be one of the required columns in the table yaml.

        If `identifier` parameter is missing or set to `None`,
        it will be set as a named parameter.
        This method accepts either a list of identifiers or a single value.
        If it receives a list it will
        behave like :func:`~table.CassandraTable.among`.
        If it receives a value it will
        behave like :func:`~table.CassandraTable.equals`.

        :param identifier: identifier or list of identifiers (e.g. device_id)
        :return: self
        """
        query = self.config['query'][self._current_query]

        if 'id' not in query['required']:
            raise NotARequiredColumnError('id', self._current_query)

        name = query['required']['id']
        if identifier is None:
            replacement = f"{name}=?"
        elif isinstance(identifier, list):
            ids_str = ', '.join([cql_quote(i) for i in identifier])
            replacement = f"{name} IN ({ids_str})"
        else:
            replacement = f"{name}={cql_quote(identifier)}"

        statements = self._prepare_column_clause(name)
        self._current_statements = \
            [self._replace(s, name, replacement) for s in statements]

        return self

    def equals(self, column: str, value: Any = None):
        """Select a value to query on a specified column.

        This column must be an optional column in the table yaml.

        If `value` parameter is missing or set to `None`,
        it will be set as a named parameter with the column name.

        :param column: column name
        :param value: value
        :return: self
        """
        if value is None:
            replacement = f"{column}=?"
        else:
            replacement = f"{column}={cql_quote(value)}"

        statements = self._prepare_column_clause(column)
        self._current_statements = \
            [self._replace(s, column, replacement) for s in statements]

        return self

    def among(self, column, values: List[Any] = None):
        """Select list of values to query on a specified column.

        This column must be an optional column in the table yaml.

        If `values` parameter is missing or set to `None`,
        it will be set as a named parameter with the column name.

        :param column: column name
        :param values: list of values to filter
        :return: self
        """
        if values is None:
            replacement = f"{column} IN (?)"
        else:
            values_str = ', '.join([cql_quote(v) for v in values])
            replacement = f"{column} IN ({values_str})"

        statements = self._prepare_column_clause(column)
        self._current_statements = \
            [self._replace(s, column, replacement) for s in statements]

        return self

    def between(
        self, column: str,
        lower: int or float = None, higher: int or float = None
    ):
        """Select range of values to query on a specified column.

        This column must be an optional column in the table yaml.

        If `lower` or `higher` parameter(s) are missing or set to `None`,
        they will be set as a named parameter with the column name -
        lt_<column-name> and ht_<column-name>.

        :param column: column name
        :param lower: lower boundary
        :param higher: higher boundary
        :return: self
        """
        if lower is None:
            replacement_lower = f"{column} > ?"
        else:
            replacement_lower = f"{column} > {lower}"

        if higher is None:
            replacement_higher = f"{column} < ?"
        else:
            replacement_higher = f"{column} < {higher}"

        if self.has_split() and column in ['day', 'month', 'year']:
            split = self.config.get('split')

            lower_datetime = datetime.fromtimestamp(lower / 1000, tz=pytz.UTC)
            higher_datetime = datetime.fromtimestamp(higher / 1000, tz=pytz.UTC)

            date_list = self._calculate_table_partitions(
                self.config['split'], lower_datetime, higher_datetime
            )

            statements = []
            for statement in self._current_statements:
                lower_split = \
                    lower_datetime.strftime(self.TABLE_FORMAT[split])
                higher_split = \
                    higher_datetime.strftime(self.TABLE_FORMAT[split])

                for date in date_list:
                    date_str = date.strftime(self.TABLE_FORMAT[split])
                    if f'_{date_str}' in statement:
                        if lower_split == date_str:
                            s = self._prepare_column_clause_single(
                                statement, column
                            )
                            statement = \
                                self._replace(s, column, replacement_lower)

                        if higher_split == date_str:
                            s = self._prepare_column_clause_single(
                                statement, column
                            )
                            statement = \
                                self._replace(s, column, replacement_higher)

                statements.append(statement)

            self._current_statements = statements
        else:
            replacement = f"{replacement_lower} AND {replacement_higher}"
            statements = self._prepare_column_clause(column)
            self._current_statements = \
                [self._replace(s, column, replacement) for s in statements]

        return self

    def between_including(
        self, column: str,
        lower: int or float = None, higher: int or float = None
    ):
        """Select range of values to query on a specified column
        including the boundary.

        This column must be an optional column in the table yaml.

        If `lower` or `higher` parameter(s) are missing or set to `None`,
        they will be set as a named parameter with the column name -
        le_<column-name> and he_<column-name>.

        :param column: column name
        :param lower: lower boundary
        :param higher: higher boundary
        :return: self
        """
        if lower is None:
            replacement_lower = f"{column} >= ?"
        else:
            replacement_lower = f"{column} >= {lower}"

        if higher is None:
            replacement_higher = f"{column} <= ?"
        else:
            replacement_higher = f"{column} <= {higher}"

        if self.has_split() and column in ['day', 'month', 'year']:
            split = self.config.get('split')

            lower_datetime = datetime.fromtimestamp(lower / 1000, tz=pytz.UTC)
            higher_datetime = datetime.fromtimestamp(higher / 1000, tz=pytz.UTC)

            date_list = self._calculate_table_partitions(
                self.config['split'], lower_datetime, higher_datetime
            )

            statements = []
            for statement in self._current_statements:
                lower_split = \
                    lower_datetime.strftime(self.TABLE_FORMAT[split])
                higher_split = \
                    higher_datetime.strftime(self.TABLE_FORMAT[split])

                for date in date_list:
                    date_str = date.strftime(self.TABLE_FORMAT[split])
                    if f'_{date_str}' in statement:
                        if lower_split == date_str:
                            s = self._prepare_column_clause_single(
                                statement, column
                            )
                            statement = \
                                self._replace(s, column, replacement_lower)

                        if higher_split == date_str:
                            s = self._prepare_column_clause_single(
                                statement, column
                            )
                            statement = \
                                self._replace(s, column, replacement_higher)

                statements.append(statement)

            self._current_statements = statements
        else:
            replacement = f"{replacement_lower} AND {replacement_higher}"
            statements = self._prepare_column_clause(column)
            self._current_statements = \
                [self._replace(s, column, replacement) for s in statements]

        return self

    def lower_than(self, column: str, boundary: int or float = None):
        """Select a minimum value to query on a specified column.

        This column must be an optional column in the table yaml.

        If `boundary` parameter is missing or set to `None`,
        it will be set as a named parameter with the column name.

        :param column: column name
        :param boundary: boundary
        :return: self
        """
        if boundary is None:
            replacement = f"{column} < ?"
        else:
            replacement = f"{column} < {boundary}"

        statements = self._prepare_column_clause(column)
        self._current_statements = \
            [self._replace(s, column, replacement) for s in statements]

        return self

    def lower_or_equal_than(self, column: str, boundary: int or float = None):
        """Select a minimum value to query on a specified column.

        This column must be an optional column in the table yaml.

        If `boundary` parameter is missing or set to `None`,
        it will be set as a named parameter with the column name.

        :param column: column name
        :param boundary: boundary
        :return: self
        """
        if boundary is None:
            replacement = f"{column} <= ?"
        else:
            replacement = f"{column} <= {boundary}"

        statements = self._prepare_column_clause(column)
        self._current_statements = \
            [self._replace(s, column, replacement) for s in statements]

        return self

    def higher_than(self, column: str, boundary: int or float = None):
        """Select a maximum value to query on a specified column.

        This column must be an optional column in the table yaml.

        If `boundary` parameter is missing or set to `None`,
        it will be set as a named parameter with the column name.

        :param column: column name
        :param boundary: boundary
        :return: self
        """
        if boundary is None:
            replacement = f"{column} > ?"
        else:
            replacement = f"{column} > {boundary}"

        statements = self._prepare_column_clause(column)
        self._current_statements = \
            [self._replace(s, column, replacement) for s in statements]

        return self

    def higher_or_equal_than(self, column: str, boundary: int or float = None):
        """Select a maximum values to query on a specified column.

        This column must be an optional column in the table yaml.

        If `boundary` parameter is missing or set to `None`,
        it will be set as a named parameter with the column name.

        :param column: column name
        :param boundary: boundary
        :return: self
        """
        if boundary is None:
            replacement = f"{column} >= ?"
        else:
            replacement = f"{column} >= {boundary}"

        statements = self._prepare_column_clause(column)
        self._current_statements = \
            [self._replace(s, column, replacement) for s in statements]

        return self

    def limit(self, value: int):
        """Limit number of results returned by the query.

        :param value: number of results
        :return: self
        """
        replacement = f"LIMIT {value}"
        self._current_statements = [self._replace(s, 'limit', replacement)
                                    for s in self._current_statements]

        return self


class CassandraMaterializedView(CassandraTable):

    @property
    def query_name(self):
        """Returns the table name."""
        return self._query_name

    def __init__(
        self,
        config: dict,
        query_name: str,
        keyspace: CassandraKeyspace = None,
        cassandra_manager: CassandraManager = None
    ):
        """Cassandra materialized view constructor.

        :param config: table template configuration
        :param query_name: materialized view name
        :param keyspace: materialized view keyspace
        :param cassandra_manager: Cassandra manager (default: None)
        """
        if query_name not in config['query']:
            raise QueryNotFound(query_name)

        super().__init__(config, keyspace, cassandra_manager)

        self._query_name = query_name

    def create(
        self,
        keyspace: str or List[str] = None,
        gc_grace_seconds: int = 86400,
        if_not_exists=False,
        **kwargs
    ):
        """Create materialized view statement.

        :param keyspace: keyspace name.
            This may be an str or List[str]  (default: None)
        :param gc_grace_seconds: garbage collect grace seconds.
            Defines the time in seconds that a tombstone is kept before being
            discarded.
        :param if_not_exists: if True, creates table if it does not exist.
            When set to False and the table exists, an exception is thrown.
        :return: self
        """
        self._current_operation = 'create'
        query = self._config['query'][self.query_name]

        keyspaces = keyspace
        if keyspaces is None:
            keyspaces = [self.keyspace_name]
        elif isinstance(keyspace, str):
            keyspaces = [keyspace]

        self._current_statements = []
        for keyspace_name in keyspaces:
            statement = f"CREATE MATERIALIZED VIEW "

            if if_not_exists:
                statement += "IF NOT EXISTS "

            statement += f"{keyspace_name}.{self.name}"
            if self.has_split():
                statement += "_{date}"
            statement += \
                f"_{self.query_name} AS SELECT * FROM {keyspace_name}.{self.name}"

            if self.has_split():
                statement += "_{date}"

            required_keys = list(query['required'].values())
            if 'optional' in query:
                required_keys += query['optional']

            statement += " WHERE "
            for index, key in enumerate(required_keys):
                if index == 0:
                    statement += f"{key} IS NOT NULL "
                else:
                    statement += f"AND {key} IS NOT NULL "

            primary_keys = ', '.join([pk for _, pk in query['required'].items()])
            statement += f"PRIMARY KEY ( ({primary_keys})"

            if 'optional' in query:
                clustering_keys = ', '.join([ck for ck in query['optional']])
                statement += f", {clustering_keys} ) "
            else:
                statement += " ) "

            statement += "WITH "

            if 'order' in query:
                statement += f"CLUSTERING ORDER BY ( "
                for index, (column, order) in enumerate(query['order'].items()):
                    if index == 0:
                        statement += f"{column} {order.upper()}"
                    else:
                        statement += f", {column} {order.upper()}"

                statement += " ) AND "

            statement += f"gc_grace_seconds={gc_grace_seconds};"

            self._current_statements.append(statement)
        return self

    def drop(
        self,
        keyspace: str or List[str] = None,
        if_exists: bool = False,
        **kwargs
    ):
        """Drop materialized view statement.

        :param keyspace: keyspace name.
            This may be an str or List[str]  (default: None)
        :param if_exists: if True, drops table if it exist.
            When set to False and the table does not exist,
            an exception is thrown (default: False)
        :return: self
        """
        self._current_operation = 'drop'

        keyspaces = keyspace
        if keyspaces is None:
            keyspaces = [self.keyspace_name]
        elif isinstance(keyspace, str):
            keyspaces = [keyspace]

        self._current_statements = []
        for keyspace_name in keyspaces:
            statement = f"DROP MATERIALIZED VIEW "
            if if_exists:
                statement += "IF EXISTS "

            statement += f"{keyspace_name}.{self.name}"
            if self.has_split():
                statement += "_{date}"
            statement += f"_{self.query_name};"
            self._current_statements.append(statement)

        return self

    def insert(self, **kwargs):
        """Raises exception. Materialized views do not support inserts."""
        raise Exception("Materialized views do not support inserts.")

    def query(self, keyspace: str or List[str] = None, **kwargs):
        """Query statement.

        This method must be chained with the methods required to define
        the primary and clustering keys, and finalizing with
        :func:`~table.CassandraTable.execute`

        :param keyspace: keyspace name.
            This may be an str or List[str]  (default: None)
        :return: self
        """
        self._current_operation = 'query'
        self._current_query = self.query_name

        keyspaces = keyspace
        if keyspaces is None:
            keyspaces = [self.keyspace_name]
        elif isinstance(keyspace, str):
            keyspaces = [keyspace]

        self._current_statements = []
        for keyspace_name in keyspaces:
            statement = (
                "SELECT {columns} FROM "
                f"{keyspace_name}.{self.name}"
            )
            if self.has_split():
                statement += "_{date}"
            statement += (
                f"_{self.query_name}"
                " {where} {limit} ;"
            )
            self._current_statements.append(statement)

        return self
